// FunctionalIIC v.0.7. 20210720. by ken woo. freely treatments.
//
// FunctionalIIC v.0.6. 20210716. by ken woo. freely treatments. modified.

// ------------- begin of functionaliic.cpp

#include"functionaliic.h"

unsigned digitalReadOutputPin(unsigned pin){ // https://forum.arduino.cc/t/how-to-find-out-output-pin-state/42066
    unsigned abit = digitalPinToBitMask(pin);
    unsigned aport = digitalPinToPort(pin);

    if (!(*portModeRegister(aport)&abit)) return digitalRead(pin); // (1==output, 0==input).

    if (aport == NOT_A_PIN) return 0;
    return (*portOutputRegister(aport) & abit) ? 1 : 0;
}

#if (MY_IIC_DEBUG)
    const char* dbg_iic_stage_name(unsigned proc){
        static const char *const z[]={
            "IDLE",
            "START",
            "REP-START",
            "RETRY",
            "ADDR",
            "WRITE",
            "READ",
            "WAIT",
            "STOP",
            "RESET"
        };
        return z[proc];
    }
#endif

SRAM_ACES cIIC::sIicMasterSig cIIC::sig; // sig must be inited before all iic objects
SRAM_ACES unsigned cIIC::dynamic_obj_num; // for the number of dynamic objects presently
SRAM_ACES volatile unsigned cIIC::SCLIO=SCLIO_DEFAULT;
SRAM_ACES volatile unsigned cIIC::SDAIO=SDAIO_DEFAULT;


SRAM_ACES cIicSlave::cSlaveQueue *cIicSlave::cSlaveQueue::head;
SRAM_ACES cIicSlave::cSlaveQueue *cIicSlave::cSlaveQueue::current_node;
SRAM_ACES unsigned cIicSlave::cSlaveQueue::current_addr;

SRAM_ACES unsigned cIicSlave::current_ctrl;
SRAM_ACES unsigned cIicSlave::current_stat;
SRAM_ACES unsigned cIicSlave::current_data;
SRAM_ACES cIicSlave::cSlaveQueue cIicSlave::sla_pool;




// -------------------- the following snippet is for esp8266 hardware timer1 triggering iic instead of polling.
// the cIIC::isr_iicproc() is used and IIC_FREQ_USED should be defined.
// and the SetHwTimer1(wrap_isr) sould be invoked.


#define FRC1_ENABLE_TIMER   (BIT7)
#define FRC1_AUTO_LOAD      (BIT6)

// note, since 1us=10x0.1us; 1us has 10 counts, for div16 @160MHz.
// and unfortunately, the timer-triggered method should not run as fast as 400k,
// instead, 15k is properly the upper bound.
// note note, if result iic freq is discrepancy,
// the clock is not attained to 160M or the 4us payload affects,
// or the watchdog and the like routines affect, or the slave gpio-int servicing latency.
#define IIC_FREQ_400K       (25/2) // the /2 since half-cycle
#define IIC_FREQ_100K       (100/2-40) // the 40, int-payload
#define IIC_FREQ_25K        (400/2-40)
#define IIC_FREQ_15K        (666/2-40)
#define IIC_FREQ_1K         (10000/2-40)
#define IIC_FREQ_200        (50000/2-40)

#define IIC_FREQ_USED       IIC_FREQ_15K

typedef enum {
    DIVDED_BY_1     = 0,        // timer clock
    DIVDED_BY_16    = 4,        // divided by 16
    DIVDED_BY_256   = 8,        // divided by 256
} time_predived_mode;

typedef enum {                  // timer interrupt mode
    TM_LEVEL_INT    = 1,        // level interrupt
    TM_EDGE_INT     = 0,        // edge interrupt
} time_int_mode;

typedef enum {
    FRC1_SOURCE     = 0,
    NMI_SOURCE      = 1,
} frc1_timer_source_type;

bool SetHwTimer1(void (*main_isr)()){
    if (!((RTC_REG_READ(FRC1_CTRL_ADDRESS))&FRC1_ENABLE_TIMER)){ // timer is disabled
        ETS_FRC_TIMER1_INTR_ATTACH(main_isr, NULL);
        /////ETS_FRC_TIMER1_NMI_INTR_ATTACH(main_isr);
        RTC_REG_WRITE((FRC1_CTRL_ADDRESS), (FRC1_ENABLE_TIMER/* | FRC1_AUTO_LOAD*/ | DIVDED_BY_16 | TM_EDGE_INT));
        TM1_EDGE_INT_ENABLE();
        ETS_FRC1_INTR_ENABLE();

        RTC_REG_WRITE((FRC1_LOAD_ADDRESS), IIC_FREQ_USED);
        return true;
    }
    return false;
}

void StopHwTimer1(){
    /////RTC_REG_WRITE((FRC1_CTRL_ADDRESS), ((RTC_REG_READ(FRC1_CTRL_ADDRESS))&~(FRC1_ENABLE_TIMER)));
    RTC_REG_WRITE((FRC1_CTRL_ADDRESS), (/*FRC1_ENABLE_TIMER | FRC1_AUTO_LOAD | */DIVDED_BY_16 | TM_EDGE_INT));
    TM1_EDGE_INT_DISABLE();
};

void ResumeHwTimer1(){

    RTC_REG_WRITE((FRC1_LOAD_ADDRESS), IIC_FREQ_USED);

    /////RTC_REG_WRITE((FRC1_CTRL_ADDRESS), ((RTC_REG_READ(FRC1_CTRL_ADDRESS))|(FRC1_ENABLE_TIMER)));
    TM1_EDGE_INT_ENABLE();
    RTC_REG_WRITE((FRC1_CTRL_ADDRESS), (FRC1_ENABLE_TIMER/* | FRC1_AUTO_LOAD*/ | DIVDED_BY_16 | TM_EDGE_INT));
};

SRAM_ACES void wrap_isr(){ // manual reload timer counter
    cIIC::isr_iicproc();
    RTC_REG_WRITE((FRC1_LOAD_ADDRESS), IIC_FREQ_USED);
}


// ------------- end of functionaliic.cpp
